@@ -3,10 +3,12 @@ from .base_miner import BaseMiner
 
 class ActivationMiner(BaseMiner):
     def __init__(
-            self, display_name, size, embeddings_field, select_filters,
-            select_limit=None, select_order_by=None, select_desc=None, 
-            select_reference_filters=None, select_reference_limit=None,select_reference_order_by=None, select_reference_desc=None,
-            skip_caching=False):
+            self, display_name, size, select_filters, model_name=None,
+            embeddings_field='embeddings',
+            select_limit=None, select_order_by=None, select_desc=None,
+            select_reference_filters=None, select_reference_limit=None,
+            select_reference_order_by=None, select_reference_desc=None,
+            skip_caching=True):
         """
         Activation miner
         Will perform a AL query based on activation
@@ -14,7 +16,7 @@ class ActivationMiner(BaseMiner):
         Parameters:
             display_name: name to be displayed in Dioptra
             size: number of datapoints to query
-            embeddings_field: embedding fields to run the analysis on. Could be 'embeddings' 'prediction.embeddings'
+            embeddings_field: embedding fields to run the analysis on. Could be 'embeddings' or 'logits'
             select_filters: dioptra style filters to select the data to be queried from
             select_limit: limit to selected the data
             select_order_by: field to use to sort the data to control how limit is performed
@@ -34,6 +36,7 @@ class ActivationMiner(BaseMiner):
                 'strategy': 'ACTIVATION',
                 'size': size,
                 'embeddings_field': embeddings_field,
+                'model_name': model_name,
                 'select': {
                     'filters': select_filters,
                     **({'limit': select_limit} if select_limit is not None else {}),
