@@ -38,7 +38,7 @@ class BaseMiner():
 
         return r.json()
 
-    def get_results(self):
+    def get_results(self, retry_on_empty_success=True):
         sleepTimeSecs = 1
         totalSleepTimeSecs = 0
 
@@ -51,11 +51,19 @@ class BaseMiner():
                 'x-api-key': self.api_key
             })
             r.raise_for_status()
-            task = r.json().get('task', {})
-            status = task.get('status', None)
+            response_json = r.json()
+            task = response_json['task']
+            status = task['status']
 
             if status == 'SUCCESS':
-                return task.get('result', [])
+                results =  task['result']
+                # TODO: remove this.
+                if len(results) == 0 and retry_on_empty_success:
+                    time.sleep(2)
+                    return self.get_results(retry_on_empty_success=False)
+                else:
+                    return results
+
             elif status == 'FAILURE' or status == 'REVOKED':
                 raise RuntimeError('Miner failed')
             else:
